# -*- coding: utf-8 -*-
#
# Copyright (c) nexB Inc. and others. All rights reserved.
# ScanCode is a trademark of nexB Inc.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/scancode-toolkit for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import os
import re

from collections import Counter
from functools import partial

from commoncode.text import as_unicode
from typecode import get_type

"""
Extract plain text from HTML, XML and related angular markup-like files and texts.
"""

# Tracing flags
TRACE = False or os.environ.get("SCANCODE_DEBUG_TEXT_ANALYSIS", False)


# Tracing flags
def logger_debug(*args):
    pass


if TRACE:
    import logging
    import sys

    logger = logging.getLogger(__name__)
    logging.basicConfig(stream=sys.stdout)
    logger.setLevel(logging.DEBUG)

    def logger_debug(*args):
        return logger.debug(" ".join(isinstance(a, str) and a or repr(a) for a in args))

extensions = (
    ".html",
    ".htm",
    ".php",
    ".phps",
    ".jsp",
    ".jspx",
    ".xml",
    ".pom",
)


def is_markup(location):
    """
    Return True is the file at `location` is some kind of markup, such as HTML,
    XML, PHP, etc.
    """
    T = get_type(location)

    # do not care for small files
    if T.size < 64:
        return False

    if not T.is_text:
        return False

    if location.endswith(extensions):
        return True

    with open(location, "rb") as f:
        start = as_unicode(f.read(1024))

    return is_markup_text(start)


def is_markup_text(text):
    """
    Return True if ``text`` is likely markup.
    """
    if text.startswith("<"):
        return True

    # count whitespaces
    no_spaces = "".join(text.split())

    # count opening and closing tags_count
    counts = Counter(c for c in no_spaces if c in "<>")

    if not all(c in counts for c in "<>"):
        return False

    if not all(counts.values()):
        return False

    # ~ 5 percent of tag <> markers measn we have tags
    has_tags = sum(counts.values()) / len(no_spaces) > 0.05

    # check if we have some significant proportion of tag-like characters
    open_close = counts[">"] / counts["<"]
    # ratio of open to close tags should approach 1: accept a 20% drift
    balanced = abs(1 - open_close) < 0.2
    return has_tags and balanced


def is_kept_tag(
    token,
    markup_markers,
    kept_markup,
):
    """
    Return True if a tag should be kept, base on its tag name or content.
    Always keep debian-style legacy <s> tags and digit-only tags
    """
    tlow = token.lower()
    if not tlow.startswith(markup_markers):
        return True

    tclean = tlow.strip("<>/").strip()
    return (
        any(kp in tclean for kp in kept_markup)
        or tlow in kept_markup
        or starts_or_ends_with_digit(tclean)
        or is_debian_tag(tlow)
    )


def starts_or_ends_with_digit(s):
    """
    Return True if a string ``s`` starts or ends with a digit. Works with empty strings.

    >>> starts_or_ends_with_digit("foo")
    False
    >>> starts_or_ends_with_digit("")
    False
    >>> starts_or_ends_with_digit(None)
    False
    >>> starts_or_ends_with_digit("foo3")
    True
    >>> starts_or_ends_with_digit("1foo3")
    True
    >>> starts_or_ends_with_digit("4foo")
    True
    >>> starts_or_ends_with_digit("3")
    True
    """
    # first and last character (works even if the string is empty)
    return bool(s and (s[:1].isdigit() or s[-1:].isdigit()))


def get_demarkuped_text(text, splitter, keeper):
    """
    Return text stripped from markup.
    Use ``splitter`` callable to split text in tokens.
    Use ``keeper`` is a callable to tell if we should keep a tag.
    """

    tokens = splitter(text)
    if TRACE:
        logger_debug(f"get_demarkuped_text: {text!r}")
        logger_debug(f"get_demarkuped_text: tokens: {tokens}")

    demarkuped = []
    demarkuped_append = demarkuped.append

    for token in tokens:
        if keeper(token):
            if token == ">":
                token = "> "
            demarkuped_append(token)
        else:
            demarkuped_append(" ")

    return " ".join("".join(demarkuped).split())


def is_debian_tag(t):
    """
    <s> are from legacy Debian copyright files.
    >>> is_debian_tag("<s>")
    True
    >>> is_debian_tag("</s>")
    True
    """
    return t in ("<s>", "</s>", "<s/>")


"""
Find start and closing tags or the first white space whichever comes first or entities.
This regex is such that ' '.join(tags.split(a))==a
"""
split_on_tags_and_entities = re.compile(
    r"("
    r"</?[^\s></]+(?:>"
    r"|"
    r"\s)?"
    r"|"
    r"&[^\s&]+;"
    r"|"
    r"href"
    r"|"
    r"['\"]?/>"
    r"|"
    r"/>"
    r")",
    re.IGNORECASE,
).split

MARKUP_MARKERS = (
    "<",
    ">",
    "/>",
    '"/>',
    "'/>",
    "&",
    "href",
)

KEPT_MARKUP = (
    "lic",
    "copy",
    "auth",
    "contr",
    # URLs
    "www",
    "http",
    # legal
    "leg",
    # encoded copyright signs
    "@",
    "169",
    "a9",
    # in <red hat inc>
    "red",
    "inc",
)

KEEPER = partial(is_kept_tag, markup_markers=MARKUP_MARKERS, kept_markup=KEPT_MARKUP)


def demarkup_text(text):
    """
    Return text lightly stripped from markup used in text preparation.
    """
    return get_demarkuped_text(text, splitter=split_on_tags_and_entities, keeper=KEEPER)


def demarkup(location, stripper=demarkup_text):
    """
    Return an iterator of unicode text lines for the file at `location` lightly
    stripping markup if the file is some kind of markup, such as HTML, XML, PHP,
    etc. The whitespaces are collapsed to one space.

    Use the ``stripper`` callable, one of demarkup_text or strip_markup_text.
    """
    from textcode.analysis import unicode_text_lines

    for line in unicode_text_lines(location):
        if TRACE:
            logger_debug(f"demarkup: {line} : demarked: {demarkup(line)}")
        yield stripper(line)


# ## Old style stripper
# this catches tags but not does not remove the text inside tags
remove_tags_legacy = re.compile(
    r"<"
    r"[(-\-)\?\!\%\/]?"
    r"[a-gi-vx-zA-GI-VX-Z][a-zA-Z#\"\=\s\.\;\:\%\&?!,\+\*\-_\/]*"
    r"[a-zA-Z0-9#\"\=\s\.\;\:\%\&?!,\+\*\-_\/]+"
    r"\/?>",
    re.MULTILINE | re.UNICODE,
).sub


def strip_markup_text_legacy(text):
    """
    Strip markup tags from ``text`` using legacy approach.
    """
    return remove_tags_legacy(" ", text).strip()


"""
Split text on tags start and end
"""
split_on_tags = re.compile(
    r"("
        # a tag
        # URL
        r"<https?://[^<>\"\']+>"
        r"|<www[^<>\"\']+>"
        r"|< */? *[a-z]+[a-z0-9@\-\._\+]* */? *>?"
        # emails
        r"|mailto:"
        r"|>"
        r"| "
    r")",
    re.IGNORECASE,
).split


def strip_known_markup_from_text(text):
    """
    Str_ip markup tags from ``text`` using a list of tags
    """
    return get_demarkuped_text(text, splitter=split_on_tags, keeper=keep_tag)


ALL_TAGS = frozenset(
    [
        "<a",
        "<a ",
        "<a>",
        "</a",
        "<a/",
        "<a/>",
        "<a/>",
        "</a>",
        "a/>",
        "/a>",
        "a>",
        " a>",
        " a>",
        "<abbr",
        "<abbr ",
        "<abbr>",
        "</abbr",
        "<abbr/",
        "<abbr/>",
        "</abbr>",
        "abbr/>",
        "/abbr>",
        "abbr>",
        " abbr>",
        "<acronym",
        "<acronym ",
        "<acronym>",
        "</acronym",
        "<acronym/",
        "<acronym/>",
        "</acronym>",
        "acronym/>",
        "/acronym>",
        "acronym>",
        " acronym>",
        "<address",
        "<address ",
        "<address>",
        "</address",
        "<address/",
        "<address/>",
        "</address>",
        "address/>",
        "/address>",
        "address>",
        " address>",
        "<applet",
        "<applet ",
        "<applet>",
        "</applet",
        "<applet/",
        "<applet/>",
        "</applet>",
        "applet/>",
        "/applet>",
        "applet>",
        " applet>",
        "<area",
        "<area ",
        "<area>",
        "</area",
        "<area/",
        "<area/>",
        "</area>",
        "area/>",
        "/area>",
        "area>",
        " area>",
        "<article",
        "<article ",
        "<article>",
        "</article",
        "<article/",
        "<article/>",
        "</article>",
        "article/>",
        "/article>",
        "article>",
        " article>",
        "<aside",
        "<aside ",
        "<aside>",
        "</aside",
        "<aside/",
        "<aside/>",
        "</aside>",
        "aside/>",
        "/aside>",
        "aside>",
        " aside>",
        "<audio",
        "<audio ",
        "<audio>",
        "</audio",
        "<audio/",
        "<audio/>",
        "</audio>",
        "audio/>",
        "/audio>",
        "audio>",
        " audio>",
        "<b",
        "<b ",
        "<b>",
        "</b",
        "<b/",
        "<b/>",
        "</b>",
        "b/>",
        "/b>",
        "b>",
        " b>",
        "<base",
        "<base ",
        "<base>",
        "</base",
        "<base/",
        "<base/>",
        "</base>",
        "base/>",
        "/base>",
        "base>",
        " base>",
        "<basefont",
        "<basefont ",
        "<basefont>",
        "</basefont",
        "<basefont/",
        "<basefont/>",
        "</basefont>",
        "basefont/>",
        "/basefont>",
        "basefont>",
        " basefont>",
        "<bdi",
        "<bdi ",
        "<bdi>",
        "</bdi",
        "<bdi/",
        "<bdi/>",
        "</bdi>",
        "bdi/>",
        "/bdi>",
        "bdi>",
        " bdi>",
        "<bdo",
        "<bdo ",
        "<bdo>",
        "</bdo",
        "<bdo/",
        "<bdo/>",
        "</bdo>",
        "bdo/>",
        "/bdo>",
        "bdo>",
        " bdo>",
        "<bgsound",
        "<bgsound ",
        "<bgsound>",
        "</bgsound",
        "<bgsound/",
        "<bgsound/>",
        "</bgsound>",
        "bgsound/>",
        "/bgsound>",
        "bgsound>",
        " bgsound>",
        "<big",
        "<big ",
        "<big>",
        "</big",
        "<big/",
        "<big/>",
        "</big>",
        "big/>",
        "/big>",
        "big>",
        " big>",
        "<blink",
        "<blink ",
        "<blink>",
        "</blink",
        "<blink/",
        "<blink/>",
        "</blink>",
        "blink/>",
        "/blink>",
        "blink>",
        " blink>",
        "<blockquote",
        "<blockquote ",
        "<blockquote>",
        "</blockquote",
        "<blockquote/",
        "<blockquote/>",
        "</blockquote>",
        "blockquote/>",
        "/blockquote>",
        "blockquote>",
        " blockquote>",
        "<body",
        "<body ",
        "<body>",
        "</body",
        "<body/",
        "<body/>",
        "</body>",
        "body/>",
        "/body>",
        "body>",
        " body>",
        "<br",
        "<br ",
        "<br>",
        "</br",
        "<br/",
        "<br/>",
        "</br>",
        "br/>",
        "/br>",
        "br>",
        " br>",
        "<button",
        "<button ",
        "<button>",
        "</button",
        "<button/",
        "<button/>",
        "</button>",
        "button/>",
        "/button>",
        "button>",
        " button>",
        "<canvas",
        "<canvas ",
        "<canvas>",
        "</canvas",
        "<canvas/",
        "<canvas/>",
        "</canvas>",
        "canvas/>",
        "/canvas>",
        "canvas>",
        " canvas>",
        "<caption",
        "<caption ",
        "<caption>",
        "</caption",
        "<caption/",
        "<caption/>",
        "</caption>",
        "caption/>",
        "/caption>",
        "caption>",
        " caption>",
        "<center",
        "<center ",
        "<center>",
        "</center",
        "<center/",
        "<center/>",
        "</center>",
        "center/>",
        "/center>",
        "center>",
        " center>",
        "<cite",
        "<cite ",
        "<cite>",
        "</cite",
        "<cite/",
        "<cite/>",
        "</cite>",
        "cite/>",
        "/cite>",
        "cite>",
        " cite>",
        "<code",
        "<code ",
        "<code>",
        "</code",
        "<code/",
        "<code/>",
        "</code>",
        "code/>",
        "/code>",
        "code>",
        " code>",
        "<col",
        "<col ",
        "<col>",
        "</col",
        "<col/",
        "<col/>",
        "</col>",
        "col/>",
        "/col>",
        "col>",
        " col>",
        "<colgroup",
        "<colgroup ",
        "<colgroup>",
        "</colgroup",
        "<colgroup/",
        "<colgroup/>",
        "</colgroup>",
        "colgroup/>",
        "/colgroup>",
        "colgroup>",
        " colgroup>",
        "<command",
        "<command ",
        "<command>",
        "</command",
        "<command/",
        "<command/>",
        "</command>",
        "command/>",
        "/command>",
        "command>",
        " command>",
        "<content",
        "<content ",
        "<content>",
        "</content",
        "<content/",
        "<content/>",
        "</content>",
        "content/>",
        "/content>",
        "content>",
        " content>",
        "<data",
        "<data ",
        "<data>",
        "</data",
        "<data/",
        "<data/>",
        "</data>",
        "data/>",
        "/data>",
        "data>",
        " data>",
        "<datalist",
        "<datalist ",
        "<datalist>",
        "</datalist",
        "<datalist/",
        "<datalist/>",
        "</datalist>",
        "datalist/>",
        "/datalist>",
        "datalist>",
        " datalist>",
        "<dd",
        "<dd ",
        "<dd>",
        "</dd",
        "<dd/",
        "<dd/>",
        "</dd>",
        "dd/>",
        "/dd>",
        "dd>",
        " dd>",
        "<del",
        "<del ",
        "<del>",
        "</del",
        "<del/",
        "<del/>",
        "</del>",
        "del/>",
        "/del>",
        "del>",
        " del>",
        "<details",
        "<details ",
        "<details>",
        "</details",
        "<details/",
        "<details/>",
        "</details>",
        "details/>",
        "/details>",
        "details>",
        " details>",
        "<dfn",
        "<dfn ",
        "<dfn>",
        "</dfn",
        "<dfn/",
        "<dfn/>",
        "</dfn>",
        "dfn/>",
        "/dfn>",
        "dfn>",
        " dfn>",
        "<dialog",
        "<dialog ",
        "<dialog>",
        "</dialog",
        "<dialog/",
        "<dialog/>",
        "</dialog>",
        "dialog/>",
        "/dialog>",
        "dialog>",
        " dialog>",
        "<dir",
        "<dir ",
        "<dir>",
        "</dir",
        "<dir/",
        "<dir/>",
        "</dir>",
        "dir/>",
        "/dir>",
        "dir>",
        " dir>",
        "<div",
        "<div ",
        "<div>",
        "</div",
        "<div/",
        "<div/>",
        "</div>",
        "div/>",
        "/div>",
        "div>",
        " div>",
        "<dl",
        "<dl ",
        "<dl>",
        "</dl",
        "<dl/",
        "<dl/>",
        "</dl>",
        "dl/>",
        "/dl>",
        "dl>",
        " dl>",
        "<dt",
        "<dt ",
        "<dt>",
        "</dt",
        "<dt/",
        "<dt/>",
        "</dt>",
        "dt/>",
        "/dt>",
        "dt>",
        " dt>",
        "<element",
        "<element ",
        "<element>",
        "</element",
        "<element/",
        "<element/>",
        "</element>",
        "element/>",
        "/element>",
        "element>",
        " element>",
        "<em",
        "<em ",
        "<em>",
        "</em",
        "<em/",
        "<em/>",
        "</em>",
        "em/>",
        "/em>",
        "em>",
        " em>",
        "<embed",
        "<embed ",
        "<embed>",
        "</embed",
        "<embed/",
        "<embed/>",
        "</embed>",
        "embed/>",
        "/embed>",
        "embed>",
        " embed>",
        "<fieldset",
        "<fieldset ",
        "<fieldset>",
        "</fieldset",
        "<fieldset/",
        "<fieldset/>",
        "</fieldset>",
        "fieldset/>",
        "/fieldset>",
        "fieldset>",
        " fieldset>",
        "<figcaption",
        "<figcaption ",
        "<figcaption>",
        "</figcaption",
        "<figcaption/",
        "<figcaption/>",
        "</figcaption>",
        "figcaption/>",
        "/figcaption>",
        "figcaption>",
        " figcaption>",
        "<figure",
        "<figure ",
        "<figure>",
        "</figure",
        "<figure/",
        "<figure/>",
        "</figure>",
        "figure/>",
        "/figure>",
        "figure>",
        " figure>",
        "<font",
        "<font ",
        "<font>",
        "</font",
        "<font/",
        "<font/>",
        "</font>",
        "font/>",
        "/font>",
        "font>",
        " font>",
        "<footer",
        "<footer ",
        "<footer>",
        "</footer",
        "<footer/",
        "<footer/>",
        "</footer>",
        "footer/>",
        "/footer>",
        "footer>",
        " footer>",
        "<form",
        "<form ",
        "<form>",
        "</form",
        "<form/",
        "<form/>",
        "</form>",
        "form/>",
        "/form>",
        "form>",
        " form>",
        "<frame",
        "<frame ",
        "<frame>",
        "</frame",
        "<frame/",
        "<frame/>",
        "</frame>",
        "frame/>",
        "/frame>",
        "frame>",
        " frame>",
        "<frameset",
        "<frameset ",
        "<frameset>",
        "</frameset",
        "<frameset/",
        "<frameset/>",
        "</frameset>",
        "frameset/>",
        "/frameset>",
        "frameset>",
        " frameset>",
        "<h1",
        "<h1 ",
        "<h1>",
        "</h1",
        "<h1/",
        "<h1/>",
        "</h1>",
        "h1/>",
        "/h1>",
        "h1>",
        " h1>",
        "<h2",
        "<h2 ",
        "<h2>",
        "</h2",
        "<h2/",
        "<h2/>",
        "</h2>",
        "h2/>",
        "/h2>",
        "h2>",
        " h2>",
        "<h3",
        "<h3 ",
        "<h3>",
        "</h3",
        "<h3/",
        "<h3/>",
        "</h3>",
        "h3/>",
        "/h3>",
        "h3>",
        " h3>",
        "<h4",
        "<h4 ",
        "<h4>",
        "</h4",
        "<h4/",
        "<h4/>",
        "</h4>",
        "h4/>",
        "/h4>",
        "h4>",
        " h4>",
        "<h5",
        "<h5 ",
        "<h5>",
        "</h5",
        "<h5/",
        "<h5/>",
        "</h5>",
        "h5/>",
        "/h5>",
        "h5>",
        " h5>",
        "<h6",
        "<h6 ",
        "<h6>",
        "</h6",
        "<h6/",
        "<h6/>",
        "</h6>",
        "h6/>",
        "/h6>",
        "h6>",
        " h6>",
        "<head",
        "<head ",
        "<head>",
        "</head",
        "<head/",
        "<head/>",
        "</head>",
        "head/>",
        "/head>",
        "head>",
        " head>",
        "<header",
        "<header ",
        "<header>",
        "</header",
        "<header/",
        "<header/>",
        "</header>",
        "header/>",
        "/header>",
        "header>",
        " header>",
        "<hgroup",
        "<hgroup ",
        "<hgroup>",
        "</hgroup",
        "<hgroup/",
        "<hgroup/>",
        "</hgroup>",
        "hgroup/>",
        "/hgroup>",
        "hgroup>",
        " hgroup>",
        "<hr",
        "<hr ",
        "<hr>",
        "</hr",
        "<hr/",
        "<hr/>",
        "</hr>",
        "hr/>",
        "/hr>",
        "hr>",
        " hr>",
        "<html",
        "<html ",
        "<html>",
        "</html",
        "<html/",
        "<html/>",
        "</html>",
        "html/>",
        "/html>",
        "html>",
        " html>",
        "<i",
        "<i ",
        "<i>",
        "</i",
        "<i/",
        "<i/>",
        "</i>",
        "i/>",
        "/i>",
        "i>",
        " i>",
        "<iframe",
        "<iframe ",
        "<iframe>",
        "</iframe",
        "<iframe/",
        "<iframe/>",
        "</iframe>",
        "iframe/>",
        "/iframe>",
        "iframe>",
        " iframe>",
        "<image",
        "<image ",
        "<image>",
        "</image",
        "<image/",
        "<image/>",
        "</image>",
        "image/>",
        "/image>",
        "image>",
        " image>",
        "<img",
        "<img ",
        "<img>",
        "</img",
        "<img/",
        "<img/>",
        "</img>",
        "img/>",
        "/img>",
        "img>",
        " img>",
        "<input",
        "<input ",
        "<input>",
        "</input",
        "<input/",
        "<input/>",
        "</input>",
        "input/>",
        "/input>",
        "input>",
        " input>",
        "<ins",
        "<ins ",
        "<ins>",
        "</ins",
        "<ins/",
        "<ins/>",
        "</ins>",
        "ins/>",
        "/ins>",
        "ins>",
        " ins>",
        "<isindex",
        "<isindex ",
        "<isindex>",
        "</isindex",
        "<isindex/",
        "<isindex/>",
        "</isindex>",
        "isindex/>",
        "/isindex>",
        "isindex>",
        " isindex>",
        "<kbd",
        "<kbd ",
        "<kbd>",
        "</kbd",
        "<kbd/",
        "<kbd/>",
        "</kbd>",
        "kbd/>",
        "/kbd>",
        "kbd>",
        " kbd>",
        "<keygen",
        "<keygen ",
        "<keygen>",
        "</keygen",
        "<keygen/",
        "<keygen/>",
        "</keygen>",
        "keygen/>",
        "/keygen>",
        "keygen>",
        " keygen>",
        "<label",
        "<label ",
        "<label>",
        "</label",
        "<label/",
        "<label/>",
        "</label>",
        "label/>",
        "/label>",
        "label>",
        " label>",
        "<legend",
        "<legend ",
        "<legend>",
        "</legend",
        "<legend/",
        "<legend/>",
        "</legend>",
        "legend/>",
        "/legend>",
        "legend>",
        " legend>",
        "<li",
        "<li ",
        "<li>",
        "</li",
        "<li/",
        "<li/>",
        "</li>",
        "li/>",
        "/li>",
        "li>",
        " li>",
        "<link",
        "<link ",
        "<link>",
        "</link",
        "<link/",
        "<link/>",
        "</link>",
        "link/>",
        "/link>",
        "link>",
        " link>",
        "<listing",
        "<listing ",
        "<listing>",
        "</listing",
        "<listing/",
        "<listing/>",
        "</listing>",
        "listing/>",
        "/listing>",
        "listing>",
        " listing>",
        "<main",
        "<main ",
        "<main>",
        "</main",
        "<main/",
        "<main/>",
        "</main>",
        "main/>",
        "/main>",
        "main>",
        " main>",
        "<map",
        "<map ",
        "<map>",
        "</map",
        "<map/",
        "<map/>",
        "</map>",
        "map/>",
        "/map>",
        "map>",
        " map>",
        "<mark",
        "<mark ",
        "<mark>",
        "</mark",
        "<mark/",
        "<mark/>",
        "</mark>",
        "mark/>",
        "/mark>",
        "mark>",
        " mark>",
        "<marquee",
        "<marquee ",
        "<marquee>",
        "</marquee",
        "<marquee/",
        "<marquee/>",
        "</marquee>",
        "marquee/>",
        "/marquee>",
        "marquee>",
        " marquee>",
        "<mathml",
        "<mathml ",
        "<mathml>",
        "</mathml",
        "<mathml/",
        "<mathml/>",
        "</mathml>",
        "mathml/>",
        "/mathml>",
        "mathml>",
        " mathml>",
        "<menu",
        "<menu ",
        "<menu>",
        "</menu",
        "<menu/",
        "<menu/>",
        "</menu>",
        "menu/>",
        "/menu>",
        "menu>",
        " menu>",
        "<menuitem",
        "<menuitem ",
        "<menuitem>",
        "</menuitem",
        "<menuitem/",
        "<menuitem/>",
        "</menuitem>",
        "menuitem/>",
        "/menuitem>",
        "menuitem>",
        " menuitem>",
        "<meta",
        "<meta ",
        "<meta>",
        "</meta",
        "<meta/",
        "<meta/>",
        "</meta>",
        "meta/>",
        "/meta>",
        "meta>",
        " meta>",
        "<meter",
        "<meter ",
        "<meter>",
        "</meter",
        "<meter/",
        "<meter/>",
        "</meter>",
        "meter/>",
        "/meter>",
        "meter>",
        " meter>",
        "<multicol",
        "<multicol ",
        "<multicol>",
        "</multicol",
        "<multicol/",
        "<multicol/>",
        "</multicol>",
        "multicol/>",
        "/multicol>",
        "multicol>",
        " multicol>",
        "<nav",
        "<nav ",
        "<nav>",
        "</nav",
        "<nav/",
        "<nav/>",
        "</nav>",
        "nav/>",
        "/nav>",
        "nav>",
        " nav>",
        "<nobr",
        "<nobr ",
        "<nobr>",
        "</nobr",
        "<nobr/",
        "<nobr/>",
        "</nobr>",
        "nobr/>",
        "/nobr>",
        "nobr>",
        " nobr>",
        "<noembed",
        "<noembed ",
        "<noembed>",
        "</noembed",
        "<noembed/",
        "<noembed/>",
        "</noembed>",
        "noembed/>",
        "/noembed>",
        "noembed>",
        " noembed>",
        "<noframes",
        "<noframes ",
        "<noframes>",
        "</noframes",
        "<noframes/",
        "<noframes/>",
        "</noframes>",
        "noframes/>",
        "/noframes>",
        "noframes>",
        " noframes>",
        "<noscript",
        "<noscript ",
        "<noscript>",
        "</noscript",
        "<noscript/",
        "<noscript/>",
        "</noscript>",
        "noscript/>",
        "/noscript>",
        "noscript>",
        " noscript>",
        "<object",
        "<object ",
        "<object>",
        "</object",
        "<object/",
        "<object/>",
        "</object>",
        "object/>",
        "/object>",
        "object>",
        " object>",
        "<ol",
        "<ol ",
        "<ol>",
        "</ol",
        "<ol/",
        "<ol/>",
        "</ol>",
        "ol/>",
        "/ol>",
        "ol>",
        " ol>",
        "<optgroup",
        "<optgroup ",
        "<optgroup>",
        "</optgroup",
        "<optgroup/",
        "<optgroup/>",
        "</optgroup>",
        "optgroup/>",
        "/optgroup>",
        "optgroup>",
        " optgroup>",
        "<option",
        "<option ",
        "<option>",
        "</option",
        "<option/",
        "<option/>",
        "</option>",
        "option/>",
        "/option>",
        "option>",
        " option>",
        "<output",
        "<output ",
        "<output>",
        "</output",
        "<output/",
        "<output/>",
        "</output>",
        "output/>",
        "/output>",
        "output>",
        " output>",
        # "<p>",
        "<param",
        "<param ",
        "<param>",
        "</param",
        "<param/",
        "<param/>",
        "</param>",
        "param/>",
        "/param>",
        "param>",
        " param>",
        "<picture",
        "<picture ",
        "<picture>",
        "</picture",
        "<picture/",
        "<picture/>",
        "</picture>",
        "picture/>",
        "/picture>",
        "picture>",
        " picture>",
        "<plaintext",
        "<plaintext ",
        "<plaintext>",
        "</plaintext",
        "<plaintext/",
        "<plaintext/>",
        "</plaintext>",
        "plaintext/>",
        "/plaintext>",
        "plaintext>",
        " plaintext>",
        "<pre",
        "<pre ",
        "<pre>",
        "</pre",
        "<pre/",
        "<pre/>",
        "</pre>",
        "pre/>",
        "/pre>",
        "pre>",
        " pre>",
        "<progress",
        "<progress ",
        "<progress>",
        "</progress",
        "<progress/",
        "<progress/>",
        "</progress>",
        "progress/>",
        "/progress>",
        "progress>",
        " progress>",
        "<q",
        "<q ",
        "<q>",
        "</q",
        "<q/",
        "<q/>",
        "</q>",
        "q/>",
        "/q>",
        "q>",
        " q>",
        "<rp",
        "<rp ",
        "<rp>",
        "</rp",
        "<rp/",
        "<rp/>",
        "</rp>",
        "rp/>",
        "/rp>",
        "rp>",
        " rp>",
        "<rt",
        "<rt ",
        "<rt>",
        "</rt",
        "<rt/",
        "<rt/>",
        "</rt>",
        "rt/>",
        "/rt>",
        "rt>",
        " rt>",
        "<ruby",
        "<ruby ",
        "<ruby>",
        "</ruby",
        "<ruby/",
        "<ruby/>",
        "</ruby>",
        "ruby/>",
        "/ruby>",
        "ruby>",
        " ruby>",
        "<s>",
        "<s/>",
        "</s>",
        "<samp",
        "<samp ",
        "<samp>",
        "</samp",
        "<samp/",
        "<samp/>",
        "</samp>",
        "samp/>",
        "/samp>",
        "samp>",
        " samp>",
        "<script",
        "<script ",
        "<script>",
        "</script",
        "<script/",
        "<script/>",
        "</script>",
        "script/>",
        "/script>",
        "script>",
        " script>",
        "<section",
        "<section ",
        "<section>",
        "</section",
        "<section/",
        "<section/>",
        "</section>",
        "section/>",
        "/section>",
        "section>",
        " section>",
        "<select",
        "<select ",
        "<select>",
        "</select",
        "<select/",
        "<select/>",
        "</select>",
        "select/>",
        "/select>",
        "select>",
        " select>",
        "<shadow",
        "<shadow ",
        "<shadow>",
        "</shadow",
        "<shadow/",
        "<shadow/>",
        "</shadow>",
        "shadow/>",
        "/shadow>",
        "shadow>",
        " shadow>",
        "<small",
        "<small ",
        "<small>",
        "</small",
        "<small/",
        "<small/>",
        "</small>",
        "small/>",
        "/small>",
        "small>",
        " small>",
        "<source",
        "<source ",
        "<source>",
        "</source",
        "<source/",
        "<source/>",
        "</source>",
        "source/>",
        "/source>",
        "source>",
        " source>",
        "<spacer",
        "<spacer ",
        "<spacer>",
        "</spacer",
        "<spacer/",
        "<spacer/>",
        "</spacer>",
        "spacer/>",
        "/spacer>",
        "spacer>",
        " spacer>",
        "<span",
        "<span ",
        "<span>",
        "</span",
        "<span/",
        "<span/>",
        "</span>",
        "span/>",
        "/span>",
        "span>",
        " span>",
        "<strike",
        "<strike ",
        "<strike>",
        "</strike",
        "<strike/",
        "<strike/>",
        "</strike>",
        "strike/>",
        "/strike>",
        "strike>",
        " strike>",
        "<strong",
        "<strong ",
        "<strong>",
        "</strong",
        "<strong/",
        "<strong/>",
        "</strong>",
        "strong/>",
        "/strong>",
        "strong>",
        " strong>",
        "<style",
        "<style ",
        "<style>",
        "</style",
        "<style/",
        "<style/>",
        "</style>",
        "style/>",
        "/style>",
        "style>",
        " style>",
        "<sub",
        "<sub ",
        "<sub>",
        "</sub",
        "<sub/",
        "<sub/>",
        "</sub>",
        "sub/>",
        "/sub>",
        "sub>",
        " sub>",
        "<summary",
        "<summary ",
        "<summary>",
        "</summary",
        "<summary/",
        "<summary/>",
        "</summary>",
        "summary/>",
        "/summary>",
        "summary>",
        " summary>",
        "<sup",
        "<sup ",
        "<sup>",
        "</sup",
        "<sup/",
        "<sup/>",
        "</sup>",
        "sup/>",
        "/sup>",
        "sup>",
        " sup>",
        "<svg",
        "<svg ",
        "<svg>",
        "</svg",
        "<svg/",
        "<svg/>",
        "</svg>",
        "svg/>",
        "/svg>",
        "svg>",
        " svg>",
        "<table",
        "<table ",
        "<table>",
        "</table",
        "<table/",
        "<table/>",
        "</table>",
        "table/>",
        "/table>",
        "table>",
        " table>",
        "<tbody",
        "<tbody ",
        "<tbody>",
        "</tbody",
        "<tbody/",
        "<tbody/>",
        "</tbody>",
        "tbody/>",
        "/tbody>",
        "tbody>",
        " tbody>",
        "<td",
        "<td ",
        "<td>",
        "</td",
        "<td/",
        "<td/>",
        "</td>",
        "td/>",
        "/td>",
        "td>",
        " td>",
        "<template",
        "<template ",
        "<template>",
        "</template",
        "<template/",
        "<template/>",
        "</template>",
        "template/>",
        "/template>",
        "template>",
        " template>",
        "<textarea",
        "<textarea ",
        "<textarea>",
        "</textarea",
        "<textarea/",
        "<textarea/>",
        "</textarea>",
        "textarea/>",
        "/textarea>",
        "textarea>",
        " textarea>",
        "<tfoot",
        "<tfoot ",
        "<tfoot>",
        "</tfoot",
        "<tfoot/",
        "<tfoot/>",
        "</tfoot>",
        "tfoot/>",
        "/tfoot>",
        "tfoot>",
        " tfoot>",
        "<th",
        "<th ",
        "<th>",
        "</th",
        "<th/",
        "<th/>",
        "</th>",
        "th/>",
        "/th>",
        "th>",
        " th>",
        "<thead",
        "<thead ",
        "<thead>",
        "</thead",
        "<thead/",
        "<thead/>",
        "</thead>",
        "thead/>",
        "/thead>",
        "thead>",
        " thead>",
        "<time",
        "<time ",
        "<time>",
        "</time",
        "<time/",
        "<time/>",
        "</time>",
        "time/>",
        "/time>",
        "time>",
        " time>",
        "<title",
        "<title ",
        "<title>",
        "</title",
        "<title/",
        "<title/>",
        "</title>",
        "title/>",
        "/title>",
        "title>",
        " title>",
        "<tr",
        "<tr ",
        "<tr>",
        "</tr",
        "<tr/",
        "<tr/>",
        "</tr>",
        "tr/>",
        "/tr>",
        "tr>",
        " tr>",
        "<track",
        "<track ",
        "<track>",
        "</track",
        "<track/",
        "<track/>",
        "</track>",
        "track/>",
        "/track>",
        "track>",
        " track>",
        "<tt",
        "<tt ",
        "<tt>",
        "</tt",
        "<tt/",
        "<tt/>",
        "</tt>",
        "tt/>",
        "/tt>",
        "tt>",
        " tt>",
        "<u",
        "<u ",
        "<u>",
        "</u",
        "<u/",
        "<u/>",
        "</u>",
        "u/>",
        "/u>",
        "u>",
        " u>",
        "<ul",
        "<ul ",
        "<ul>",
        "</ul",
        "<ul/",
        "<ul/>",
        "</ul>",
        "ul/>",
        "/ul>",
        "ul>",
        " ul>",
        "<var",
        "<var ",
        "<var>",
        "</var",
        "<var/",
        "<var/>",
        "</var>",
        "var/>",
        "/var>",
        "var>",
        " var>",
        "<video",
        "<video ",
        "<video>",
        "</video",
        "<video/",
        "<video/>",
        "</video>",
        "video/>",
        "/video>",
        "video>",
        " video>",
        "<wbr",
        "<wbr ",
        "<wbr>",
        "</wbr",
        "<wbr/",
        "<wbr/>",
        "</wbr>",
        "wbr/>",
        "/wbr>",
        "wbr>",
        " wbr>",
        "<xlink",
        "<xlink ",
        "<xlink>",
        "</xlink",
        "<xlink/",
        "<xlink/>",
        "</xlink>",
        "xlink/>",
        "/xlink>",
        "xlink>",
        " xlink>",
        "<xml",
        "<xml ",
        "<xml>",
        "</xml",
        "<xml/",
        "<xml/>",
        "</xml>",
        "xml/>",
        "/xml>",
        "xml>",
        " xml>",
        "<xml",
        "<xml ",
        "<xml>",
        "</xml",
        "<xml/",
        "<xml/>",
        "</xml>",
        "xml/>",
        "/xml>",
        "xml>",
        " xml>",
        "<xmlns",
        "<xmlns ",
        "<xmlns>",
        "</xmlns",
        "<xmlns/",
        "<xmlns/>",
        "</xmlns>",
        "xmlns/>",
        "/xmlns>",
        "xmlns>",
        " xmlns>",
        "<xmp",
        "<xmp ",
        "<xmp>",
        "</xmp",
        "<xmp/",
        "<xmp/>",
        "</xmp>",
        "xmp/>",
        "/xmp>",
        "xmp>",
        " xmp>",
        # not XML/HTML
        "<year>",
        "<name>",
        "</name>",
        "<url>",
        "</url>",
        "<date-of-document>",
        # common XML namespaces
        "http://www.w3.org/1998/math/mathml",
        "http://www.w3.org/1999/xhtml",
        "http://www.w3.org/1999/xlink",
        "http://www.w3.org/2000/svg",
        "http://www.w3.org/2000/xmlns/",
        "http://www.w3.org/xml/1998/namespace",
    ]
)


SKIP_ATTRIBUTES = (
    "href=",
    "HREF=",
    "class=",
    "CLASS=",
    "width=",
    "@end",
    "@group",
    "mailto:",
    "xmlns=",
    "xmlns:",
    "xml:",
    "lang=",
    "<Windows",
    "<windows",
)


KEEP_MARKERS = (
    "copyright",
    "author",
    "legal",
)


def keep_tag(token, skips_tags=ALL_TAGS, skip_attributes=SKIP_ATTRIBUTES, kept_tags=KEEP_MARKERS):
    """
    Return True if a tag should be kept, base on a list of tag name or content.
    """
    tlow = token.lower()

    if any(k in tlow for k in kept_tags):
        return True

    if token.startswith(skip_attributes):
        return False

    if tlow in skips_tags or tlow == ">":
        return False

    return True
